#!/usr/bin/perl

### mknorm V 0.3; copyright 2003 by Stefan Heimann
## input:
# 0 : scanner-shortcut
# 1 : reportfile of scanner
# 2 : normfile


if ($#ARGV < 2) {# proof number of given parameters
    print("ERROR: missing parameter!!! \n");
    exit 1;
}


## include modules ##

use FileHandle; # perl module for file handles
use normfile_methods; # methods for the normfile, for instance read out the normfile
use texttools; # tools for texthandling
use filehandling; # tools for file handling
use mknorm_fmw_key; # key module for F-MakroW
use mknorm_fpr_key; # key module for F-Prot
use mknorm_vsc_key; # key module for McAfee VirusScan
use mknorm_rfwcrc_key; # key module for RFW-checksum-list
use mknorm_rfwlog_key; # key module for RFW-douple/nullbyte-sample-list
use mknorm_extconv_key; # key module for extension-conventions of the av-producer
use mknorm_file_key; # key module for file.pl, an file format analyser

## Global variables ##

%samples; # key: samplepath; value: identification-hash

$scanner = $ARGV[0]; # scanner-shortcut to invoke special scanner-module
$RepFile = $ARGV[1]; # input-reportfile
$NormFile = $ARGV[2]; # input-existing-normfile

$NewNormFile = $NormFile; # output-new-normfile replacing existing normfile

my @RepFiletmp = split(/\./,$RepFile); # temp var to split filename from extension
$DebugFile = $RepFiletmp[0]."_mknorm\.dbg"; # file for debugging
$NotNormFile = $RepFiletmp[0]."_notnorm\.lst"; # list of not in norm samplepath or identification
$UnrelIdentifFile = $RepFiletmp[0]."_unrelidentif\.lst"; # list of samples unreliable detected
$NoIdentifFile = $RepFiletmp[0]."_noidentif\.lst"; # list of samples unreliable detected without identification
$SBOFile = $RepFiletmp[0]."_sbo\.lst"; # list of samples reported OK
$RestFile = $RepFiletmp[0]."_rest\.lst"; # rest of lines in the reportfile


## open / backup / create files ##

$DebugFH = new FileHandle "> $DebugFile" or do {
        print("ERROR creating $DebugFile : $!\n");
        closeall(); # close all opened files
        exit 1;
        };

$NormFile = &filehandling::backup_file($NormFile); # backup normfile, if it exists
if ($NormFile ne "#new#") { # if a normfile exists then read out samples and identifications
	%samples = &normfile_methods::read_normfile($NormFile);
}


## call here main-norming-routines ##

print ($DebugFH "\$scanner mknorm : $scanner\n");
print ($DebugFH "\$RepFile mknorm : $RepFile\n");
print ($DebugFH "\$NormFile mknorm : $NormFile\n");

norming(); # norm given report, of given scanner

writenorm(); # write normed sample and identification(s) to NewNormFile

closeall();

&filehandling::rm_nullbytefiles($NotNormFile,$UnrelIdentifFile,$NoIdentifFile,$SBOFile,$RestFile); # remove output files with null byte


### Sub-Routines ###


sub norming {

        print($DebugFH "mknorm_$scanner invoked\n");

        $RepFH = new FileHandle "< $RepFile" or do {
                print("ERROR opening $RepFile : $!\n");
                closeall();
                exit 1;
                };

        $NotNormFH = new FileHandle "> $NotNormFile" or do {
                print("ERROR creating $NotNormFile : $!\n");
                closeall();
                exit 1;
                };

        $UnrelIdentifFH = new FileHandle "> $UnrelIdentifFile" or do {
                print("ERROR creating $UnrelIdentifFile : $!\n");
                closeall();
                exit 1;
                };

        $NoIdentifFH = new FileHandle "> $NoIdentifFile" or do {
                print("ERROR creating $NoIdentifFile : $!\n");
                closeall();
                exit 1;
                };

        $SBOFH = new FileHandle "> $SBOFile" or do {
                print("ERROR creating $SBOFile : $!\n");
                closeall();
                exit 1;
                };

        $RestFH = new FileHandle "> $RestFile" or do {
                print("ERROR creating $RestFile : $!\n");
                closeall();
                exit 1;
                };

        local $line; # current report-file-line
        local $lastline = ""; # last report-file-line (may be used in future for scanners, which identification is seperated into more then one line)
        local $variant_seperator = "-"; # seperator for generic identification with multiple variants
        local $dropper_variant_mark = "_drv_"; # mark for extra dropper variant info
        local $infective_length_mark = "_"; # mark for infective length
        local $devolution_mark = "_d_"; # mark for (d)evolutions
        local $language_mark = "_l_"; # mark for languages
        local $addition_mark = "__"; # mark for additions
        local $extension_mark = "."; # mark for extensions
        local $sample; # path of a sample
        local $type; # which type of malware : script, macro, (boot), file, exot
        local $class; # which class of malware : virus, trojan, worm, ...
        local $platform; # identified platform : w97m, w32, vbs, ...
        local $family; # identified family
        local $variant; # identified variant
        local $infective_length; # instad of a variant : length of infective code in byte
        local $devolution; # not a particular variant, but differend because of a programming error
        local $language; # regional language : de, fr, tw, ...
        local $addition; # like @MM
        local $extension; # extension for file format : exe, mdb, vbs, doc, ...
        local $filelength; # filelength of the sample (used by RFW)
        local $crc1; # first crc-checksum (from RFW)
        local $crc2; # second crc-checksum (from RFW)

        while (defined($line = <$RepFH>)) {
                $line = &texttools::shchommp($line); # cut of empty characters from end of line, for instance \n; letters in upper case
                print($DebugFH "\nnorming $line\n");
                $variant = "a"; # default variant, if no one is given in identification
                $infective_length = $infective_length_mark;
                $devolution = $devolution_mark;
                $language = $language_mark;
                $addition = $addition_mark;
                $extension = $extension_mark;
                $filelength = "";
                $crc1 = "";
                $crc2 = "";
                if ($scanner eq "fmw") { # howto: dynamischer Aufruf der Module ohne Abfrage eines Scannerkrzels
                        &mknorm_fmw_key::parsing(); # key module for the scanner F-MacroW
                        #$lastline = $line; # see remark above for $lastline
                }
                elsif ($scanner eq "fpr") {
                        &mknorm_fpr_key::parsing(); # key module for the scanner F-Prot
                }
                elsif ($scanner eq "vsc") {
                        &mknorm_vsc_key::parsing(); # key module for the scanner VirusScan from McAfee
                }
                elsif ($scanner eq "rfwcrc") {
                        &mknorm_rfwcrc_key::parsing(); # key module for RFW-checksum-list
                }
                elsif ($scanner eq "rfwlog") {
                        &mknorm_rfwlog_key::parsing(); # key module for RFW-douple/nullbyte-sample-list
                }
                elsif ($scanner eq "extconv") {
                        &mknorm_extconv_key::parsing(); # key module for extension-conventions of the av-producer
                }
                elsif ($scanner eq "file") {
                        &mknorm_file_key::parsing(); # key module for file.pl, an file format analyser
                }
                else {
                        print ("ERROR : Scanner $scanner unknown : Module not included or wrong scannershortcut!\n");
                        closeall();
                        exit 1;
                }
        }
        close($RepFH);
        close($NotNormFH);
        close($UnrelIdentifFH);
        close($NoIdentifFH);
        close($SBOFH);
        close($RestFH);
}


sub compare_alter { # compare with existing informations from normfile and alter if new

        my $identification = $type.":\/".$class."\/".$platform."\/".$family."\/";
        if ($variant ne "") {
                $identification .= $variant;
        }
        else {
                $identification .= $infective_length;
        }
        if ($devolution ne "") {
                $identification .= $devolution;
        }
        if ($language ne "") {
                $identification .= $language;
        }
        if ($addition ne "") {
                $identification .= $addition;
        }
        $identification .= "\/".$scanner.$extension."|".$filelength.";".$crc1.";".$crc2;

        $identification = uc($identification);
        $sample = uc($sample);

        if (defined($::samples{$sample})) { # sample exists in normfile
                if (defined(${$::samples{$sample}}{$identification})) { # identification exists
                        print($DebugFH "  sample-identification does already exist : \n\t".$sample." ".$identification."\n");
                }
                else { # new identification
                        ${$::samples{$sample}}{$identification} = 0;
                        print($DebugFH "    new identification found for $sample ");
                        print($DebugFH "$identification\n");
                }
        }
        else { # new sample
                ${$::samples{$sample}}{$identification} = 0;
                print($DebugFH "    new entry $sample $identification found\n");
        }
}


sub writenorm { # write samples and identifications into new normfile

        print($DebugFH "write out into new normfile $NewNormFile\n");

        $NewNormFH = new FileHandle "> $NewNormFile" or do { # create new normfile or exit on error
                print("ERROR creating $NewNormFile : $!\n");
                closeall();
                exit 1;
                };

        my $sample; # current sample to write out
        my $identifikation; # identification for current sample
        foreach $sample (keys(%samples)) {
                print($NewNormFH $sample."\n");
                foreach $identification (keys(%{$samples{$sample}})) {
                        print($NewNormFH "\t".$identification."\n");
                }
                print ($NewNormFH "\n"); # blank line for better human readability
                #print($DebugFH "sample and identification(s) for $sample printed\n");
        }
        close($NewNormFH);
}


sub closeall { # close all files
	close($DebugFH);
	close($NormFH);
        close($NewNormFH);
        close($RepFH);
	close($NotNormFH);
        close($UnrelIdentifFH);
        close($NoIdentifFH);
        close($SBOFH);
        close($RestFH);
}