#!/usr/bin/perl

### mktestbed V 0.1; copyright 2003 by Stefan Heimann
## input:
# 0 : identiffile (infile)
# 1 : batfile (main outfile)
# 2 : destination directory
# [3 : (optional)
#
#       "-move" : move samples into testbeds; default: copy samples only
# ]


if ($#ARGV < 2) {# proof number of given parameters
    print("ERROR: missing parameter!!! \n");
    exit 1;
}


## include modules ##

use FileHandle; # perl module for file handles
use normfile_methods; # methods for the NormFile, for instance read out the NormFile
use identiffile_methods; # methods for the IdentifFile, for instance read out the IdentifFile
use texttools; # tools for texthandling
use filehandling; # tools for file handling

## Global variables ##

$variant_seperator = "-"; # seperator for generic identification with multiple variants
$dropper_variant_mark = "_drv_"; # mark for extra dropper variant info
$infective_length_mark = "_"; # mark for infective length
$devolution_mark = "_d_"; # mark for (d)evolutions
$language_mark = "_l_"; # mark for languages
$addition_mark = "__"; # mark for additions
$extension_mark = "."; # mark for extensions

$move = 0; # flag for moving samples into testbed; default: no

%sample_identifications = (); # sample identifications hash; key: unified (relative) samplepath; value: single identification
%testbedpaths = (); # all paths used in all testbeds

$IdentifFile = $ARGV[0]; # file of full single sample identifications
$MkTestbedFile = $ARGV[1]; # file for all testbeds: first it is easier to make a Batch-File
$NewMkTestbedFile = $MkTestbedFile;

if (defined($ARGV[2])) { # get options, if exist
        my $argvstr = "";
        for($zaehler = 0; $zaehler < @ARGV; $zaehler++) {
                $argvstr .= " ".$ARGV[$zaehler];
        }

        if ($argvstr =~ /-move/i) {
                $move = 1; # set move flag; default: do not move samples
        }
}

my @IdentifFiletmp1 = split(/\./,$IdentifFile); # temp var to split filename from extension
my @IdentifFiletmp2 = split(/_/,$IdentifFiletmp1[0]); # temp var to split "_identif" from filename, if exists
$DebugFile = $IdentifFiletmp2[0]."_mktestbed\.dbg"; # file for debugging
$NotNormFile = $IdentifFiletmp2[0]."_mktestbed_notnorm\.lst"; # list of out of norm sample identifications in IdentifFile
$NotTestbedFile = $IdentifFiletmp2[0]."_mktestbed_nottestbed\.lst"; # list of sample identifications, which could not be put into any testbed
$RestFile = $IdentifFiletmp2[0]."_mktestbed_identifrest\.lst"; # list of sample identifications, which were not worked up, for instance because sample not in -workpath

$logfile = $IdentifFiletmp2[0]."_mkidentif\.log"; # file for debugging


## open / backup / create files ##


$DebugFH = new FileHandle "> $DebugFile" or do {
        print("ERROR creating $DebugFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$RestFH = new FileHandle "> $RestFile" or do {
        print("ERROR creating $RestFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$NotNormFH = new FileHandle "> $NotNormFile" or do {
        print("ERROR creating $NotNormFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$NotTestbedFH = new FileHandle "> $NotTestbedFile" or do {
        print("ERROR creating $NotTestbedFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };

$MkTestbedFile = &filehandling::backup_file($MkTestbedFile); # backup mktestbedfile, if it exists

$NewMkTestbedFH = new FileHandle "> $NewMkTestbedFile" or do {
        print("ERROR creating $NewMkTestbedFile : $!\n");
        closeall(); # close all opened files
        exit 1; # stop script
        };


## call here main-routines ##


print ($DebugFH "IdentifFile mktestbed : $IdentifFile\n");
print ($DebugFH "MkTestbedFile mktestbed : $NewMkTestbedFile\n");
print ($DebugFH "workpath : $workpath\n");
print ($DebugFH "move samples into testbed : $move\n");

%sample_identifications = &identiffile_methods::read_identiffile($IdentifFile); # read out identiffile

make_testbeds();

closeall();

&filehandling::rm_nullbytefiles($NotNormFile,$RestFile,$NotTestbedFile); # remove output files with null byte


### Sub-Routines ###


sub make_testbeds { # make the testbeds

        print($DebugFH "making testbeds ...\n");

        my $sample; # path of a sample

        my $sample_testbedpath;
        foreach $sample (keys(%sample_identifications)) { # iterate all samples and put them into the testbeds
                print ($DebugFH "\nsample : ".$sample."\n");
                $sample_testbedpath = get_testbedpath($sample,$sample_identifications{$sample},0);
                if (not $sample_testbedpath =~ /ERROR/i) {
                        print ($NewMkTestbedFH "echo put '$sample' -> '$sample_testbedpath'>>$logfile\n");
                        print ($NewMkTestbedFH "echo f | xcopy $sample $sample_testbedpath>>$logfile\n");
                        if ($move == 1) { # move samples
                                print ($NewMkTestbedFH "echo delete '$sample'>>$logfile\n");
                                print ($NewMkTestbedFH "del '$sample'>>$logfile\n");
                        }
                        print ($NewMkTestbedFH "\n");
                }
                elsif ($sample_testbedpath eq "identif_ERROR") { # ERROR in Identification # not in use
                        print ($NotNormFH "$sample\n");
                        print ($NotNormFH "\tNOTNORM: ".to_identificationstring($sample_identifications{$sample})."\n\n");
                }
                elsif ($sample_testbedpath eq "number_ERROR") { # sample-number is too big
                        print ($NotTestbedFH "number_ERROR:\n$sample\n");
                        print ($NotTestbedFH "\t".to_identificationstring($sample_identifications{$sample})."\n\n");
                }
                else { # not in use
                        print ($RestFH "$sample\n");
                        print ($RestFH "\t".to_identificationstring($sample_identifications{$sample})."\n\n");
                }
        }##end: foreach $sample (keys(%sample_identifications))
}##end: sub make_testbeds


sub get_testbedpath {
#in: sample, identification array, number
#out: path to put sample into

        my $sample = $_[0];
        my @identif_parts = @{$_[1]};

        my $testbedpath = "";

        my $pretestbedpath = ""; # testbedpath till number
        my $number = 0;
        my $testbedpath_ERROR = 0; # Exception-ERROR-code:
        # 0: no Errors
        # 1: Identification ERROR: Identification is out of norm # not in use
        # 2: number ERROR: number is >= 1000 : to big => sample-filename would be out of testbed norm (old 8.3 filesystemformat)


        print ($DebugFH "\nget_testbedpath:\n\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                        #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}

        #              type                    class                  platform               family
        $pretestbedpath = $identif_parts[0]."\\".$identif_parts[1]."\\".$identif_parts[2]."\\".$identif_parts[3]."\\";
        #   variant
        if ($identif_parts[4] ne "") {
                $pretestbedpath .= $identif_parts[4];
        }
        #      infective_length
        if ($identif_parts[5] ne "") {
                $pretestbedpath .= $infective_length_mark.$identif_parts[5];
        }
        #   devolution
        if ($identif_parts[6] ne "") {
                $pretestbedpath .= $devolution_mark.$identif_parts[6];
        }
        #   language
        if ($identif_parts[7] ne "") {
                $pretestbedpath .= $language_mark.$identif_parts[7];
        }
        #   addition
        if ($identif_parts[8] ne "") {
                $pretestbedpath .= $identif_parts[8];
        }
        #                       platform
        $pretestbedpath .= "\\".$identif_parts[2]."_"; # (next to append: testbed sample number)

        my $numberstring;
        while ( ( (defined ($testbedpaths{$testbedpath})) || ($testbedpath eq "") ) && ($testbedpath_ERROR == 0) ) { # find not used number for the testbed-sample

                $numberstring = leading_zeros($number,3); # add leading zeros to number, for example: "7" -> "007", "42" -> "042"

                if ( ($numberstring ne "ERROR") && ($testbedpath_ERROR == 0) ) { # adding leading zeros successed
                        #                                                 extension
                        $testbedpath = $pretestbedpath.$numberstring."_".$extension_mark.$identif_parts[9];
                        $testbedpath = uc($testbedpath);
                        print ($DebugFH "  try testbedpath: '".$testbedpath."'\n");
                }
                elsif ($numberstring eq "ERROR")  { # numberstring_ERROR
                        $testbedpath_ERROR = 2;
                }

                $number++;
        }

        if ($testbedpath_ERROR == 0) { # insert testbedpath into existing paths hash
                $testbedpaths{$testbedpath} = 1;
                print ($DebugFH "  testbedpath for sample found : '".$testbedpath."'");
        }
        elsif ($testbedpath_ERROR == 1)  { # indicate identif_ERROR to upper routine
                $testbedpath = "identif_ERROR";
        }
        elsif ($testbedpath_ERROR == 2)  { # indicate number_ERROR to upper routine
                $testbedpath = "number_ERROR";
        }

        return $testbedpath;
}##end: get_testbedpath()


sub leading_zeros {
#in: a number, numberlength

        my $number = $_[0];
        my $to_numberlength = $_[1];

        my $numberstring = $number;

        if (length($numberstring) < $to_numberlength) {
                for($counter = length($numberstring); $counter < $to_numberlength; $counter++) {
                                $numberstring = "0".$numberstring;
                }
                print ($DebugFH "    add leading zeros : $number -> $numberstring\n");
        }
        elsif (length($numberstring) > $to_numberlength) {
                print ($DebugFH "    ERROR : given number is too big\n");
                $numberstring = "ERROR";
        }

        return $numberstring;
}


sub to_identificationparts { # converts an identification string into an array
#in : identification string
#out : identification array

        my $identificationstr = $_[0]; # string to extract each identification part from

        my @identification_array; # array : [0]: $scanner|"NOTNORM" ; [1]: array of identification parts (without $scanner)

        my $type; # which type of malware : script, macro, (boot), file, exot
        my $class; # which class of malware : virus, trojan, worm, ...
        my $platform; # identified platform : w97m, w32, vbs, ...
        my $family; # identified family
        my $variant; # identified variant
        my $infective_length; # instad of a variant : length of infective code in byte
        my $devolution; # not a particular variant, but differend because of a programming error
        my $language; # regional language : de, fr, tw, ...
        my $addition; # like @MM
        my $scanner; # shortcut for the scanner, the identification belongs to
        my $extension; # extension for file format : exe, mdb, vbs, doc, ...
        my $filelength; # filelength of the sample (used by RFW)
        my $crc1; # first crc-checksum (from RFW)
        my $crc2; # second crc-checksum (from RFW)

        #print ($DebugFH "\nto_identificationparts:\n\n");
        #closures                  1        2       3       4       567          8                                9               10     11             12            13                     14     15             16     17                    18              19    20              21       (20+)23           24     25
        if ($identificationstr =~ /(\w+)?:\/(\w+)?\/(\w+)?\/(\S+)?\/((([a-z]{1,3}($variant_seperator[a-z]{1,3})*)?($devolution_mark(\d+))?($language_mark([a-z]{2}))?)|($infective_length_mark(\w+)))?($addition_mark(\w+))*($dropper_variant_mark([a-z]{1,3}))?\/(\w+)?($extension_mark(\w{1,3}))?\|((\d+\.)?\d+)?;(\w+)?;(\w+)?/i) {
                $type = uc($1);
                $class = uc($2);
                $platform = uc($3);
                $family = uc($4);
                $variant = uc($7);
                $infective_length = uc($14);
                $devolution = uc($10);
                $language = uc($12);
                $addition = uc($15);
                $addition .= uc($17);
                $scanner = uc($19);
                $extension = uc($21);
                $filelength = uc($22);
                $crc1 = uc($24);
                $crc2 = uc($25);

                $identification_array[0] = $scanner; # use while identifiing; while writing out: = "";
                $identification_array[1] = [$type,$class,$platform,$family,$variant,$infective_length,$devolution,$language,$addition,$extension,$filelength,$crc1,$crc2];
        }
        else { # identification is out of norm
                $identification_array[0] = "NOTNORM"; # indicate, that identification is out of norm
        }

        return @identification_array;
}


sub to_identificationstring { # converts an array of identification parts into an identification string
#in : identification array
#out : identification string

        # index:   0     1      2         3       4        5                 6           7         8         9          10          11    12
        # fields: [type, class, platform, family, variant, infective_length, devolution, language, addition, extension, filelength, crc1, crc2]
        my @identif_parts = @{$_[0]}; # identification array

        my $identification;

        print ($DebugFH "\nto_identificationstring:\n\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}
        #                 type                    class                  platform               family
        $identification = $identif_parts[0].":\/".$identif_parts[1]."\/".$identif_parts[2]."\/".$identif_parts[3]."\/";
        #   variant
        if ($identif_parts[4] ne "") {
                $identification .= $identif_parts[4];
        }
        #      infective_length
        elsif ($identif_parts[5] ne "") {
                $identification .= $infective_length_mark.$identif_parts[5];
        }
        #   devolution
        if ($identif_parts[6] ne "") {
                $identification .= $devolution_mark.$identif_parts[6];
        }
        #   language
        if ($identif_parts[7] ne "") {
                $identification .= $language_mark.$identif_parts[7];
        }
        #   addition
        if ($identif_parts[8] ne "") {
                $identification .= $identif_parts[8];
        }
        #                                       extension             filelength             crc1                   crc2
        $identification .= "\/".$extension_mark.$identif_parts[9]."|".$identif_parts[10].";".$identif_parts[11].";".$identif_parts[12];

        $identification = uc($identification);

        return $identification;
}


sub proof_full_identif { # proof, that full identification is made
#in: identification parts
#out: boolean for full identification

        my @identif_parts = @{$_[0]};

        my @essential_partinds = ("1", "2", "3", "9", "4x5");

        my $full_identified = 1; # boolean for full identification

        print ($DebugFH "\nproof_full_identif:\n\n");
        #print ("proof_full_identif:\n");
        #print ($DebugFH "  \$#identif_parts: $#identif_parts\n");
        #for ($counter = 0; $counter <= $#identif_parts; $counter++) {
                #print ($DebugFH "  \$identif_parts\[$counter\]: '".$identif_parts[$counter]."'\n");
        #}

        full_identified: while (1) {
                foreach $partindex (@essential_partinds) {
                        if (($partindex =~ /^(\d+)$/i) && ($identif_parts[$1] eq "")) {
                                $full_identified = 0;
                                print ($DebugFH "proof_full_identif: not full identified (and) :  $1 \n");
                                last full_identified;
                        }
                        elsif ($partindex =~ /^(\d+)(x(\d+))+$/i) {
                                my $xor_partinds = $partindex;
                                my $info = 0;
                                while ($xor_partinds ne "") {
                                        #print ("  is defined \$xor_partinds: ".$xor_partinds."\n");
                                        if ($xor_partinds =~ /^(\d+)(x(.+))?$/i) {
                                                if ($3) { $xor_partinds = $3; } else { $xor_partinds = ""; }
                                                if (($info == 1) && ($identif_parts[$1] ne "")) {
                                                        $info = 1;
                                                        $full_identified = 0;
                                                        $xor_partinds = "";
                                                        print ($DebugFH "proof_full_identif: not full identified (xor) :  $1 : $identif_parts[$1]\n");
                                                        last full_identified;
                                                }
                                        }
                                        else {
                                             print("      ERROR: \$xor_partinds: '".$xor_partinds."'\n");
                                             print("      ERROR: \$identif_parts[$1]: '".$identif_parts[$1]."'\n");
                                        }
                                }##end: while ($xor_partinds ne "")
                        }##end: elsif "xor"
                        #else {
                             #print ("ERROR: not AND and not XOR\n");
                        #}
                }##end: foreach $partindex (@essential_partfinds)

                last full_identified;
        }##end: while "true"

        return $full_identified;
}


sub closeall { # close all files
    close($DebugFH);
    close($IdentifFH);
    close($MkTestbedFile);
    close($NewMkTestbedFile);
    close($NotTestbedFH);
    close($RestFH);
    close($NotNormFH);
}