#!/usr/bin/perl

### proof8_3.pl V 1.0.2; copyright 2003 by Stefan Heimann
## proof the content of a directory by recursive deep-search to solve problems with filesystemstructures not in 8.3-format :
#
# - replace special characters in paths with "_"
# - cut names and extensions to 8.3-format
# - numberate, if new name exists; for example "filenames.ext" to "filena_1.ext", if "filename.ext" already exists
# - delete zerobyte-files
# - delete empty directories recursive
# - give a statistic about number of "real" files and subdirectories
#
# Parameters :
#
# 0 :   Directorypath for Directory to proof
#
# [ optional:
#
#       -ps=<pathseperator>     : default is "\" for Windows-Systems
#
#       -nodemo                 : proof for problems and try to solve them; default is to proof only and write
#                                 problems to log-file
#
# ]

##to do:
#       - fullpath handling, for example "c:\dir"
#       - handling for relative path inputdir parameter, for example: "..\dir"
#       - choose-work options, for example: do only delete zerobyte-files and empty directories


use FileHandle; # perl module
use DirHandle; # perl module


if ($#ARGV < 0) {# proof number of given parameters
    print("ERROR: missing parameter!!! \n");
    exit 1;
}


## global variables ##

$inputdir = $ARGV[0]; # given dir, which content is to proof recursively

$pathseperator = '\\'; # default pathseperator (windows systems)
$demo = 1; # variable for demo mode; default : log only

%paths; # hash of existing paths in current directory for working in demo mode

if ($inputdir =~ /\Q$pathseperator\E$/i) { # cut of pathseperator at end of directory to proof
        chop($inputdir);
}

$LogFile = $inputdir."_proof\.log"; # one-line-output file of actions, that have been done
$DebugFile = $inputdir."_proof\.dbg"; # output file for debugging


## open / create files ##

$LogFH = new FileHandle "> $LogFile" or do {
        print("ERROR creating $LogFile : $!\n");
        close();
        exit 1;
        };

$DebugFH = new FileHandle "> $DebugFile" or do {
        print("ERROR creating $DebugFile : $!\n");
        closeall();
        exit 1;
        };


## option handling ##

if (defined($ARGV[1])) { # if exist, get options
        my $argstring = "";
        my $counter;
        for($counter = 1; $counter < @ARGV; $counter++) { # get options from array into a string
                $argstring .= " ".$ARGV[$counter];
        }
        print ($DebugFH "options : $argstring\n");

        if ($argstring =~ /(.*?)\s-ps=(\S{1,2})(\s?.*)/i) { # search for pathseperator option
                $pathseperator = $2; # set customized pathseperator
                $argstring = $1.$3; # reduce string of options
                #print ($DebugFH "pathseperator : $pathseperator\n");
        }
        if ($argstring =~ /(.*)\s-nodemo(\s?.*)/i) { # search for demo mode option
                $demo = 0; # set no demo mode !
                $argstring = $1.$2; # reduce string of options
                print ($DebugFH "NODEMO!\n");
        }

        if ($argstring =~ /\S/i) { # handling for unknown options
                print ($DebugFH "ERROR: unknown option(s): $argstring\n" );
                print ("ERROR: unknown option(s): $argstring\n\n" );
                closeall(); # close all opened files
                unlink($LogFile); # delete empty log file
                exit 1;
        }
        print ($DebugFH "\n\n");
}

print ($DebugFH "pathseperator : $pathseperator\n");


### main ###

my %dirstats; # hash of statistics for input directory (file-, subdir-number)
%dirstats = proof_content("  ",$inputdir); # proof content of given directory, than given directory itself
print ($LogFH "\n\n");
foreach $statistic (keys(%dirstats)) { # write statistics to LogFile
        my $intendstr = "";
        for($counter = 1; $counter < (20-length($statistic)-length($dirstats{$statistic})+2); $counter++) {
                $intendstr .= " ";
        }
        print ($LogFH "$statistic :".$intendstr.$dirstats{$statistic}."\n");
}

closeall(); # close all opened files


### subroutines ###


sub proof_content { # recursive called routine to proof directories and their content
#in0: string to indent lines in DebugFile
#in1: directorypath


        my $prestring = $_[0]."  "; # expand string to indent lines recursively
        my $proofdir = $_[1]; # directory to proof

        my %dircounts = ("files" => 0, "subdirs" => 0); # hash to count files and subdirectories

        %paths = (); # initialize name hash new for equal paths in this directory only, not in its subdirectories

        my $proofdirDH = new DirHandle $proofdir;
        opendir($proofdirDH,$proofdir) or do { # open given directory
                print("ERROR opening $proofdir : $!\n");
                closedir($proofdirDH);
                closeall();
                exit 1;
                };

        my $dirobject; # a single file or subdirectory
        my $relpath; # relative path to $dirobject from startpoint
        my %subdircounts; # a single returned %dircount of a subdirectory
        while (defined($dirobject = readdir($proofdirDH))) { # get a directory object
                $relpath = $proofdir.$pathseperator.$dirobject; # set relative path
                if (-d $relpath) { # proof subdirectory recursive
                        if (not ($dirobject =~ /\.{1,2}$/i)) { # exclude current dir entry "." and parent dir entry ".."
                                print ($DebugFH "$prestring subdir : $dirobject\n");
                                %subdircounts = proof_content($prestring,$relpath); # proof subdirectory recursively
                                if ($subdircounts{"files"} > 0) { # add statistics of subdirectory, if not empty
                                        $dircounts{"subdirs"} += $subdircounts{"subdirs"} + 1; # number of sub-subdirs plus the subdir itself
                                        $dircounts{"files"} += $subdircounts{"files"}; # add filenumber of subdir (and its subdirs)
                                }
                                elsif ($subdircounts{"subdirs"} == -1) { # subdir is empty, but unremovable
                                        print ($DebugFH "$prestring subdir '".$dirobject."' is empty but unremovable\n\n");
                                        $dircounts{"subdirs"}++;
                                }
                        }
                }
                elsif (-f $relpath) { # proof file
                        #print ($DebugFH "$prestring file : $dirobject\n");
                        if (not (proof_file($prestring,$relpath) =~ /deleted\|/i)) { # increase number of files, if file is not a zerobyte-file
                                        $dircounts{"files"}++;
                        }
                }
                print ($DebugFH "\n");
        }
        print ($DebugFH $prestring."close($proofdir)\n");
        closedir($proofdirDH); # close directory

        if ( ($dircounts{"files"} == 0) && ($dircounts{"subdirs"} == 0) ) { # remove directory, if empty
                print ($DebugFH "$prestring try to delete empty directory: $proofdir\n");
                if ($demo == 0) { # if not working in demo mode -> do remove directory
                        rmdir($proofdir) or do { # try to remove directory or handle error
                                print($DebugFH "$prestring  WARNING: Can not delete '".$proofdir."' : $!\n");
                                print($LogFH "WARNING: Can not delete '".$proofdir."' : $!\n");
                                $dircounts{"subdirs"} = -1; # indicate to upper recursion level, that empty dir still exists
                        };
                }
                if ($dircounts{"subdirs"} != -1) { print ($LogFH "delete empty directory: $proofdir\n"); }; # write successful removing to LogFile
        }
        else { # proof to rename directory itself, if necessary
                proof_rename($prestring,{get_parts($prestring,$proofdir)});
        }
        print ($DebugFH "\n\n");

        return %dircounts; # return statistics for this directory (and its subdirectories)
}


sub proof_file { # proof file for zerobyte; rename file, if necessary
#in: filepath

        my $prestring = $_[0];
        my $file = $_[1]; # path of file to proof

        my %fileparts; # hash of splited fileparts: path to directory of file, name of file, extension of file
        my $actions = "";

        $actions .= proof_del_zbfile($prestring,$file); # proof for zerobyte

        if ($actions eq "") { # if no action is reported => not a zerobyte-file
                %fileparts = get_parts($prestring,$file); # split filepath
                if (not ($fileparts{"name"} =~ /\/ERROR\//)) { # spliting OK
                        #foreach $filepart (keys(%fileparts)) {
                                #print ($DebugFH "$prestring proof_file : $filepart => ".$fileparts{$filepart}."\n");
                        #}
                        $actions .= proof_rename($prestring,{%fileparts}); # proof file for renaming
                }
                else { # spliting was not OK
                        $actions .= "ERROR|";
                        exit 1;
                }
        }

        return $actions; # return all actions done with file
}


sub proof_del_zbfile { #
#in: filepath
#out: actionsstring

        my $prestring = $_[0];
        my $file = $_[1]; # filepath

        my $actions = ""; # for logging actions

        if (-z $file) { # delete zerobyte-file
                print ($DebugFH "$prestring try to delete zerobyte file : $file\n");
                $actions .= "deleted|"; # log action
                if ($demo == 0) { # if not working in demo mode -> do delete file
                        unlink($file) or do { # try to delete file or handle error
                                print($DebugFH "$prestring  WARNING: Can not delete $file : $!\n");
                                print($LogFH "WARNING: Can not delete $file : $!\n");
                                $actions =~ s/deleted\|/deleting_failed\|/gi; # do not count file as deleted, but do not try to rename
                        };
                }
                if ($actions =~ /delete\|/i) { print ($LogFH "delete zerobyte file: $file\n"); }; # write delete action to LogFile
        }

        return $actions;
}


sub get_parts {
#in1: path of file or directory
#out: hash for splited path

        my $prestring = $_[0];
        my $path = $_[1];

        my %pathparts; # hash of parts from the path: prepath, name, extension

        if ($path =~ /^(.+\Q$pathseperator\E)?([^\Q$pathseperator\E]+)\.([^\Q$pathseperator\E\.]+)$/i) { # split path for paths with extension (mostly files)
                $pathparts{"prepath"} = $1;
                $pathparts{"name"} = $2;
                $pathparts{"extension"} = $3;
                print ($DebugFH "$prestring path (ext):\n");
                print ($DebugFH "$prestring  prepath    : ".$pathparts{"prepath"}."\n");
                print ($DebugFH "$prestring  name       : ".$pathparts{"name"}."$name\n");
                print ($DebugFH "$prestring  extension  : ".$pathparts{"extension"}."$extension\n");
        }
        elsif ($path =~ /^(.+\Q$pathseperator\E)?([^\Q$pathseperator\E]+)$/i) { # split path for paths without extension (mostly directories)
                $pathparts{"prepath"} = $1;
                $pathparts{"name"} = $2;
                $pathparts{"extension"} = "";
                print ($DebugFH "$prestring path (noext):\n");
                print ($DebugFH "$prestring  prepath    : ".$pathparts{"prepath"}."\n");
                print ($DebugFH "$prestring  name       : ".$pathparts{"name"}."$name\n");
                print ($DebugFH "$prestring  extension  : ".$pathparts{"extension"}."$extension\n");
        }
        else { # can not split path
                $pathparts{"prepath"} = "";
                $pathparts{"name"} = "\/ERROR\/";
                $pathparts{"extension"} = "";
                print ($DebugFH "$prestring ERROR: Can not split path : $path\n");
                exit 1;
        }

        return %pathparts; # return hash for splited path
}


sub proof_rename { # proof some parts of the splited path for renaming
#in2: hash of pathparts
#out: actionsstring

        my $prestring = $_[0];
        my %pathparts = %{$_[1]}; # hash for splited path

        #foreach $pathpart (keys(%pathparts)) {
                #print ($DebugFH "$prestring proof_rename : $pathpart => ".$pathparts{$pathpart}."\n");
        #}

        local $actions = ""; # to log actions
        my %new_pathparts = %pathparts; # hash for modified, splited path

        print ($DebugFH "$prestring proof_rename: name\n");
        $new_pathparts{"name"} = replace_special_chars($prestring,$new_pathparts{"name"}); # proof name for special characters and replace them
        $new_pathparts{"name"} = cut_to($prestring,$new_pathparts{"name"},8); # proof and cut name to a length of 8 characters

        if ($new_pathparts{"extension"} ne "") { # modify extension, if necessary
                print ($DebugFH "$prestring proof_rename: extension\n");
                $new_pathparts{"extension"} = replace_special_chars($prestring,$new_pathparts{"extension"}); # proof extension for special characters and replace them
                $new_pathparts{"extension"} = cut_to($prestring,$new_pathparts{"extension"},3); # proof and cut extension to a length of 3 characters
        }

        if ($actions ne "") { # rename, if any part was modified
                my $path = to_path({%pathparts}); # get a path again
                my $new_path = proof_find_newname($prestring,{%new_pathparts}); # proof and find the new path
                print ($DebugFH "$prestring try to rename : '".$path."' ---> '".$new_path."'\n");
                $actions .= "rename: >".$path."< -> >".$new_path."<|"; # log renaming action
                if ($demo == 0) { # if not working in demo mode -> do rename object
                        rename($path,$new_path) or do { # try to rename object or handle error
                                print($DebugFH "$prestring  WARNING: Can not rename '".$path."' : $!\n");
                                print($LogFH "WARNING: Can not rename '".$path."' : $!\n");
                                $actions =~ s/rename: >$path< -> >$new_path<\|//gi; # unlog renaming action
                        };
                }
                if ($actions =~ /rename:/i) { print ($LogFH "rename: '".$path."' ---> '".$new_path."'\n"); }; # write rename action to logfile
        }

        print ($DebugFH "$prestring actions: $actions\n");
        return $actions;
}


sub replace_special_chars { # replace special characters with "_"
#in: string
#out: new string

        my $prestring = $_[0];
        my $old_string = $_[1]; # string, which might contain special characters

        my $new_string = $old_string; # new string, which might contain replaced characters

        #my $special_chars = "[\!\\$\%\&\(\)\=\+\#\~\^\.\@\\[\]\{\}\\\'\` ]";
        my $special_chars = '[^a-zA-Z0-9\-_]'; # argument for special characters
        my $replace_char ='_'; # character to replace with

        print ($DebugFH "$prestring proof for special chars : $new_string\n");
        if ($new_string =~ /$special_chars/i) { # if string contains special characters
                $new_string =~ s/$special_chars/$replace_char/gi; # replace special characters
                $actions .= "replace_special_chars|";
                print ($DebugFH "$prestring '".$old_string."' --replace--> '$new_string'\n");
        }

        return $new_string; # return string with replaced characters
}


sub cut_to { # cut string to maximum length
#in1: string
#in2: maximum length
#out: new string

        my $prestring = $_[0];
        my $old_string = $_[1];
        my $max_length = $_[2]; # maximum length to cut string to

        my $new_string = $old_string; # new, cut string

        print ($DebugFH "$prestring proof for maximum length : $new_string\n");
        if (length($new_string) > $max_length) { # length > max_length => cut of tail > max_length
                $new_string = substr($new_string,0,$max_length);
                $actions .= "cut_to_$max_length|";
                print ($DebugFH "$prestring '".$old_string."' --cut--> '$new_string'\n");
        }

        return $new_string;
}


sub to_path { # rejoin the parts of a path to the path-string
#in: hash of pathparts
#out: path-string

        my %pathparts = %{$_[0]};

        my $path = $pathparts{"prepath"}.$pathparts{"name"};

        if ($pathparts{"extension"} ne "") { # join extension to path, if exists
                $path .= "\.".$pathparts{"extension"};
        }

        return $path; # return path-string
}


sub proof_find_newname { # proof and find a new name for the object, that does not exist yet
#in: hash of pathparts
#out: new path-string

        my $prestring = $_[0];
        my %pathparts = %{$_[1]}; # hash for splited path

        my %new_pathparts = %pathparts; # modified hash for splited path
        my $new_path; # new path-string

        my $newpath_ok = 0; # indicator for success of search
        my $counter = 1; # counter for $counteradd
        my $counteradd; # number of characters to replace at end of path
        while (($newpath_ok == 0) && ($counter < 9999999)) { # search for new, not existing path; <prepath>"_9999999"[.<extension>] is the limit of search
                $new_path = to_path({%new_pathparts}); # get path-string from path parts
                print($DebugFH "$prestring try $new_path\n");
                if ((-e $new_path) || (defined($paths{uc($new_path)}))) { # new path does exist; for demo mode: look in hash of existing paths
                        $counteradd = 8 - (length($counter) + 1); # replace number of length of counter-string plus one ("_") characters
                        $new_pathparts{"name"} = substr($new_pathparts{"name"},0,$counteradd)."_".$counter; # replace last characters with "_" and a number (like windows "~"<number>)
                        print($DebugFH "$prestring searching for newname; number : $counter\n");
                        $counter++; # increase counter for next try, if modified path does also exist
                        $actions .= "newname_searched|"; # log, that new path had to be modified
                }
                else { # new path does not exist
                        $newpath_ok = 1; # indicate end and success of search
                        $paths{uc($new_path)} = 0; # add new path to hash of existing paths
                        print($DebugFH "$prestring newname found : $new_path\n");
                }
        }

        if ($newpath_ok == 0) { # got out of search loop, but search was not successfull => exception handling
                $actions .= "ERROR|";
                print($DebugFH "$prestring  ERROR: Can not rename : ".to_path({%pathparts})." : rename limit reached\n");
                print($LogFH " ERROR: Can not rename : ".to_path({%pathparts})." : rename limit reached\n");
                closeall();
                exit 1; # abort on exception: a number of nearly 10.000.000 renamed objects in one directory is improbable
                        # a possibility to increase limit is to chop the path and call this method recursively
        }

        return $new_path; # return new (modified) path-string
}


sub closeall { # close all files, that might be opened
        close($DebugFH);
        close($LogFH);
}